#! /usr/local/bin/perl -w

require 5;
use strict;
use Socket;
use Date::Manip; # Available at http://www.perl.org/CPAN/CPAN.html

my ( $PRIMEDIA ) = "primedia";
my ( $CODEDB ) = "opcode";

{
    # Initialize the server socket
    my( $port ) = 10000;
    my( $tcpProtocolNumber ) = getprotobyname( 'tcp' ) || 6;
    my( $internetPackedAddress ) = 
	pack( 'Sna4x8', AF_INET(), $port, "\0\0\0\0" );
    my( $connection, $addr, $child, $af, $clientPort, $internetAddr, 
	$hostname);
    $SIG{ INT } = \&close_socket;	
    $SIG{ CHLD } = sub { wait }; # Kill zonbie!!;-)
    socket( SOCKET, AF_INET(), SOCK_STREAM(), $tcpProtocolNumber ) 
	or die( "socket: $!" );	
    print STDOUT "Port = $port\n";		
    bind( SOCKET, $internetPackedAddress )
	or die( "bind: $!" );		
    listen( SOCKET, 5 ) or die( "listen: $!" );    
    select( SOCKET ); $|=1;
    
    # server main loop
    for ( $connection=0; ;$connection++ ) {
	printf STDOUT ( "Listening for connection %d........\n", $connection );
	$addr = accept( NEWSOCKET, SOCKET ) or die( "Accept: $!" );   
	if ( ( $child = fork() ) == 0 ) {	
	    print STDOUT "Accept ok\n";	
	    ( $af, $clientPort, $internetAddr ) = unpack( 'Sna4x8', $addr );
	    $hostname = gethostbyaddr( $internetAddr, AF_INET ); 
	    print STDOUT 
		"$connection: port = $clientPort , From : $hostname\n";
	    select( NEWSOCKET ); $| = 1;

	    my( $input ) = ""; # buffer init
	    recv( NEWSOCKET, $input, 10000000, 0 ); 
	    if ( $input =~ /^INS/ ){ &insert( $input ); }      
	    elsif ( $input =~ /^DEL/ ) { &delete( $input ); }   
	    elsif ( $input =~ /^TNS/ ){ &transclude( $input ); }
	    elsif ( $input =~ /^REA/ ){ &rearrange( $input ); }
	    elsif ( $input =~ /^!INS/ ){ &send_mat( $input ); }
	    elsif ( $input =~ /^!REA/ ){ &send_swapped_ptr( $input ); }
	    
	    ### Will be disabled
	    elsif ( $input =~ /^!VCP/ ){ &v_comparing( $input ); }
	    ###
	    
	    elsif ( $input =~ /^REQ/ ){ &request( $input ); }
	    elsif ( $input =~ /^DOC/ ){ &doc( $input ); }
	    else { print STDOUT "Type Unknown.\n";}
	    close( NEWSOCKET );
	    print STDOUT "Connection closed.\n";
	    print STDOUT "\$input: $input \n";
	    exit;			
	}
    }
    close( SOCKET );
}

# Accepts the new material from a front end, 
# appends it to the primedia database,
# with appending the edit operation information to opcode database.
# Then sends the pointer and material to the front end. 
sub insert ( $input )
{			
    my( $input ) = @_;
    my( $primedia_len, $start, $len, $output ); 
    my( $owner, $doc, $time, $version, $local, $material ) 
	= &divide( $input ); 

    open( PRIMEDIA, ">> $PRIMEDIA" ) or die "Can't open \"$PRIMEDIA\"!: $!\n";
    select( PRIMEDIA ); $|=1; 
    $primedia_len = -s PRIMEDIA; 
    seek( PRIMEDIA, 0, 2 );		# seek EOF address
    print PRIMEDIA "$material";
    close( PRIMEDIA );
    $len = length( $material );	
    $start = $primedia_len; 
    $output = "[$start+$len]/$material";
#    print STDOUT "insert: \$output is $output\n";
    send( NEWSOCKET, $output, 0 );
    my( $code ) = 
	$owner."/".$doc."/".$version."/".$time."/INS/".$local."[$start+$len]";
    &opcode( $code );       
    return;
}

# Accepts the pointer from a front end, and sends back the material.
sub send_mat ( $ )
{
    my( $input ) = @_;

    my( $ptr ) = substr( $input, 5 );
    my( @ptrs ) = split( /\[/, $ptr );
    shift( @ptrs );  # discard the first undef char
    my( $outmat ) = "";
    foreach( @ptrs ){
        my( $start, $len ) = &remove_chars( $_ );
        my( $material ) = &get_material( $start, $len );
        $outmat = $outmat.$material;
    }
    my( $output ) = $outmat;
#    print STDOUT "send_mat: \$output is $output\n";
    send( NEWSOCKET, $output, 0 );
    return;
}

# Accepts the pointers, swaps them, then sends them back to the front end.  
sub send_swapped_ptr ( $ )
{
    my( $input ) = @_;

    my( $ptr ) = substr( $input, 5 );
    my( @ptrs ) = split( /\s+/, $ptr );
    my( @swapped ) = reverse( @ptrs );
    my( $swapped ) = join( "", @swapped );
#    print STDOUT "send_swapped_ptr: \$swapped is $swapped\n";
    send( NEWSOCKET, $swapped, 0 );
    return;
}

# Divides the information into pieces.
sub divide ( $ ) 
{            # ex.TNS/DocOwner/DOCNAME/time/2 A 1/3/...[.....]
    my( $input ) = @_;
    $_ = substr( $input, 4 );
    my( $seg ) = index( $_, "/" );
    my( $owner ) = substr( $_, 0, $seg );
    $_ = substr( $_, $seg+1 );
    $seg = index( $_, "/" );
    my( $doc ) = substr( $_, 0, $seg );
    $_ = substr( $_, $seg+1 );
    $seg = index( $_, "/" );
    my( $time ) = substr( $_, 0, $seg );
    $_ = substr( $_, $seg+1 );
    $seg = index( $_, "/" );
    my( $ver ) = substr( $_, 0, $seg );
    $_ = substr( $_, $seg+1 );
    $seg = index( $_, "/" );
    my( $local ) = substr( $_, 0, $seg );
    my( $info ) = substr( $_, $seg+1 );
    return( $owner, $doc, $time, $ver, $local, $info );
}

# Accepts the pointer information, 
# appending the edit operation information to opcode database.
# Then sends the pointer(s) to the front end.
sub delete ( $input )
{   
    my( $input ) = @_;
    my( $start, $len, @ptrs, $whitespace, @outptr );	
    my( $owner, $doc, $time, $version, $local, $ptr ) 
	= &divide( $input );

    @ptrs = split( /\[/, $ptr );
    shift( @ptrs );
    my( $outmat ) = "";
    foreach( @ptrs ) {
	( $start ,$len ) = &remove_chars( $_ );
	$whitespace = " " x $len;
	push( @outptr, "[$start+$len]" );
	$outmat = $outmat.$whitespace;
    }
    my( $output ) = join( "", @outptr )."/$outmat";
#   print STDOUT "delete: \$output is $output\n";
    send( NEWSOCKET, $output, 0 );
    my( $code ) = 
      $owner."/".$doc."/".$version."/".$time."/DEL/".$local.join( "", @outptr );
    &opcode( $code );
    return;
}

# Accepts the pointer information,
# appending the edit operation information to opcode database.
# Then sends the pointer and material to the front end.
sub transclude ( $input )
{		
    my( $input ) = @_;
    my( $start, $len, @ptrs, $material, @outptr );
    my( $owner, $doc, $time, $version, $local, $ptr ) 
	= &divide( $input );

    @ptrs = split( /\[/, $ptr );
    shift( @ptrs );  # discard the first undef char
    my( $outmat ) = "";
    foreach( @ptrs ) {
	( $start, $len ) = &remove_chars( $_ ); 
	$material = &get_material( $start, $len );
	push( @outptr, "[$start+$len]" );
        $outmat = $outmat.$material;   
    }
    my( $output ) = join( "", @outptr )."/$outmat";
#    print STDOUT "transclude: \$output is $output\n";
    send( NEWSOCKET, $output, 0 );
    my( $code ) = 
      $owner."/".$doc."/".$version."/".$time."/TNS/".$local.join("", @outptr);
    &opcode( $code );
    return;
}

# Accepts the pointer information,
# appending the edit operation information to opcode database.
# Then sends the pointer and material to the front end.
sub rearrange ( $input )
{
    my( $input ) = @_;
    my( $start, $len, @ptrs, @swapped, $material, @outptr ); 
    my( $owner, $doc, $time, $version, $local, $ptr ) 
	= &divide( $input ); 

    my( @before ) = split( /\*/, $ptr ); # different delimeter!
    my( $before ) = qq( @before ); # Note: 
    @swapped = reverse( @before ); 
    my( $swapped ) = join( "", @swapped ); 
    my( @newptrs ) = split( /\[/, $swapped );    
    shift( @newptrs ); # discard the first undef char
    my( $outmat ) = "";
    foreach( @newptrs ) {
	( $start, $len ) = &remove_chars( $_ );
	$material = &get_material( $start, $len );
	push( @outptr, "[$start+$len]" );
        $outmat = $outmat.$material;
    }		
    my( $output ) = join( "", @outptr )."/".$outmat;
#    print STDOUT "rearrange: \$output is $output\n";
    send( NEWSOCKET, $output, 0 );
    my( $code ) 
      = $owner."/".$doc."/".$version."/".$time."/REA/".$local.$before; 
                                          # Note: $before with space
    &opcode( $code );
    return;
}

# Seeks and reads the primedia, and returns the material.
sub get_material( $$ )
{
    my( $start, $len ) = @_; 
    my( $material ) = ""; # init buffer
    open( PRIMEDIA, $PRIMEDIA ) or die "Can't open \" $PRIMEDIA\": $!\n";
    seek( PRIMEDIA, $start, 0 );	
    read( PRIMEDIA, $material, $len ); 
    close( PRIMEDIA );
    return $material;
}			       

# Strips the [..+..] charactors from the pointer.
sub remove_chars ( $ )
{
    my( $ptr ) = @_;		
    my( $start, $len );
    my( @ptr ) = split( /\]/, $ptr );
    $_ = shift( @ptr );
    ( $start, $len ) = ( split( /\+/ ,$_ ) );         
    return ( $start, $len ); 
}

# Appends a new operation code to the opcode dababase.
sub opcode ( $ )
{
    my( $opcode ) = @_;
    open( OPCODE, ">> $CODEDB" ) or die "Can't open \"$CODEDB\"!: $!\n";
    select( OPCODE ); $|=1;
    print OPCODE "$opcode\n";
    close( OPCODE );
    return;
}

# Finds the matarial by DOCUMENT NAME.
sub request ( $ )  
{
    my( $input ) = @_;
    my( $codebuf ) = "";
    my( @send );
    $_ = substr( $input, 4 ); # input is REQ/DocOwner/DOCNAME/.....
    my( $seg ) = index( $_, "/" );
    my( $docowner ) = substr( $_, 0, $seg );
    $_ = substr( $_, $seg+1 );
    $seg = index( $_, "/" );
    my( $docmt ) = substr( $_, 0, $seg );
    my( $vers ) = substr( $_, $seg+1 );

    undef $/;
    open( CODE, $CODEDB ) or die "Can't open \"$CODEDB\": $!\n";
    read( CODE, $codebuf, -s CODE );
    close( CODE );
    my( @codes ) = split( /\n+/, $codebuf );
    my( @request ) = split( /\n+/, $vers );
    my( $ver, $each );

    foreach $ver ( @request ) {
      foreach $each ( @codes ) {
	my( $owner, $doc, $version, $time, $type, $local_ptrs )
	  = divide_code( $each );
	if ( $docmt eq $doc ) {
	  if ( $version eq $ver ) {
	    my( $ok ) = $version."/".$type."/".$local_ptrs;
	    push( @send, $ok );
	  } 
	}
      }
    }
    
    my( $send );
    if ( defined @send ) {
	@send = reverse( @send );
	$send = join( "\n", @send );
	$send = $send."\n";
    } else { $send = "Not found"; }
    send( NEWSOCKET, $send, 0 );
    return;
}

# Pulls out the information from the opcode.
sub divide_code ( $ ) 
{
    my( $input ) = @_;  # ex.DocOwner/DOCNAME/2 A 3/INS/10[......]
    my( $seg ) = index( $input, "/" );
    my( $owner ) = substr( $input, 0, $seg );
    $_ = substr( $input, $seg+1 );
    $seg = index( $_, "/" );
    my( $doc ) = substr( $_, 0, $seg );
    $_ = substr( $_, $seg+1 );
    $seg = index( $_, "/" );
    my( $ver ) = substr( $_, 0, $seg );
    $_ = substr( $_, $seg+1 );
    $seg = index( $_, "/" );
    my( $time ) = substr( $_, 0, $seg );
    $_ = substr( $_, $seg+1 );
    $seg = index( $_, "/" );
    my( $type ) = substr( $_, 0, $seg );
    my( $local_ptrs ) = substr( $_, $seg+1 );
    return( $owner, $doc, $ver, $time, $type, $local_ptrs );
}

# Finds the material by DOCUMENT OWNER NAME.
sub doc ( $input )
{
    my( $input ) = @_;
    $_ = substr( $input, 4 );
    my( $seg ) = index( $_, "/" );
    my( $docowner ) = substr( $_, 0, $seg );
    my( $docmt ) = substr( $_, $seg+1 );
    my( $codebuf ) = "";
    my( @send );

    undef $/;
    open( CODE, $CODEDB ) or die "Can't open \"$CODEDB\": $!\n";
    read( CODE, $codebuf, -s CODE );
    close( CODE );
    my( @codes ) = split( /\n+/, $codebuf );

    my( $latest ) = 0;
    my( $latest_ver );

    if ( $docowner eq "" ) {
        foreach ( @codes ) {
	  my( $owner, $doc, $version, $time, $type, $local_ptrs )
	    = divide_code( $_ );
	  if ( $docmt eq $doc ) {
	    push( @send, $version );
	    
	    $time = &ParseDate( $time );
	    if ( $time gt $latest ) {
	      $latest = $time;
	      $latest_ver = $version;
	    }
	    
	  }
        }
    } else {
      foreach ( @codes ) {
	my( $owner, $doc, $version, $time, $type, $local_ptrs )
	  = divide_code( $_ );
	if ( $docowner eq $owner ) {
	  if ( $docmt eq $doc ) {
	    push( @send, $version );
	    
	    $time = &ParseDate( $time );
	    if( $time gt $latest ) {
	      $latest = $time;
	      $latest_ver = $version;
	    }
	  }
	}
      }
    }
    my( $send ) = join( "\n", @send );
    if ( $send eq "" ) {
      $send = "Not found";
    } else {
      $send = $send."\n".$latest_ver."\n";
    }
    send( NEWSOCKET, $send, 0 );
    return;
}

# Closes socket.
sub close_socket ()
{
    close( NEWSOCKET );
    die( "Socket closed due to SIGINT.\n" );
}







